// ExactFormulae.cpp
//
// Some exact formulae.
//
// (C) Datasim Education BV 2011
//

#ifndef Exact_hpp
#define Exact_hpp

#include "GeneralInclude.hpp"

double N(double x)
{
	normal_distribution<> myNormal(0.0, 1.0); 

	return cdf(myNormal, x);
}


struct TwoFactorBasketData
{ // Payoff = max(a1S1 - a2S2,0)

	double s1, s2, rho; // Volatility and correlation
	double T;			// Expiry date
	double r;			// Interest date
	double a1, a2;		// Coefficients in payoff
	int Q1, Q2;			// Quantities of S1 and S2 to exchange
	double b1, b2;		// Cost of carry
	double q1, q2;		// Pay-out rates
	double K;			// Strike

	void print() const
	{
		cout << endl;
		cout << "Vols and correlation: " << s1 << ", " << s2 << ", " << rho << endl;
		cout << "Expiry and interest: " << T << ", " << r << endl;
		cout << "Coefficients a1 and a2 in max(a1S1 - a2S2,0): " << a1 << ", " << a2 << endl;
		cout << "Quantities Q1(S1) and Q2(S2) to exchange: " << Q1 << ", " << Q2 << endl;
		cout << "Cost of carry: " << b1 << ", " << b2 << endl;
		cout << "Dividends: " << q1 << ", " << q2 << endl;

		cout << "Modified volatility: " << sqrt(s1*s1 + s2*s2 - 2.0*rho*s1*s2);
		cout << "\nStrike: " << K << endl;

		cout << endl;
	}
};



double Margrabe(const TwoFactorBasketData& data, double S1, double S2)
{ // Call price for exchange option at the underlyings (S1_0, double S2_0)
  // Payoff is max(Q1*S1 - Q2*S2, 0)

	if (S2 == 0.0)
		return S1;

	double sig = sqrt(data.s1*data.s1 + data.s2*data.s2 - 2.0*data.rho*data.s1*data.s2); 
	double sqT = sig*sqrt(data.T);

	double d1 = (log(data.Q1*S1/(data.Q2*S2)) + (data.b1 - data.b2 + 0.5*sig*sig)*data.T) / sqT;
	double d2 = d1 - sqT;
	//cout << "d1 d2" << d1 << ", " << d2 << endl;
	//cout << "N(d1 d2)" << N(d1) << ", " << N(d2) << endl;
	return data.Q1*S1*exp((data.b1-data.r)*data.T)*N(d1) - data.Q2*S2*exp((data.b2-data.r)*data.T)*N(d2);
}


double OutperformanceOption(double S1, double S2, int cp, double b1, double b2, double v1, double v2, double X, double T, double rho, double r)
{
	if (S2 == 0.0)
		return 0.0;

	// Haug
	// F = S1 / S2 * Exp((b1 - b2 + v2 ^ 2 - rho * v1 * v2) * T)
	double F = S1/S2*exp((b1 - b2 + v2*v2 - rho*v1*v2)*T);

	//v = Sqr(v1 ^ 2 + v2 ^ 2 - 2 * rho * v1 * v2)
	double v = sqrt(v1*v1 + v2*v2 - 2.0*rho*v1*v2); 
	
	// d1 = (Log(F / X) + T * v ^ 2 / 2) / (v * Sqr(T))
	// d2 = d1 - v * Sqr(T)
	double d1 = (log(F/X) + 0.5*v*v*T) / (v*sqrt(T));
	double d2 = d1 - v*sqrt(T);

	if (cp < 0)
	{ // Put
		
		// OutPerformance = Exp(-r * T) * (X * CND(-d2) - F * CND(-d1))
		return exp(-r*T)*(X*N(-d2) - F*N(-d1));
	}
	else 
	{ // Call
		
		//OutPerformance = Exp(-r * T) * (F * CND(d1) - X * CND(d2))
		return exp(-r*T)*((F*N(d1) - X*N(d2)));
	}
}

double OutperformanceOption(const TwoFactorBasketData& data, double S1, double S2, int cp)
{
	if (S2 == 0.0)
		return 0.0;

	// Haug
	// F = S1 / S2 * Exp((b1 - b2 + v2 ^ 2 - rho * v1 * v2) * T)
	double F = S1/S2*exp((data.b1 - data.b2 + data.s2*data.s2 - data.rho*data.s1*data.s2)*data.T);

	//v = Sqr(v1 ^ 2 + v2 ^ 2 - 2 * rho * v1 * v2)
	double v = sqrt(data.s1*data.s1 + data.s2*data.s2 - 2.0*data.rho*data.s1*data.s2); 
	
	// d1 = (Log(F / X) + T * v ^ 2 / 2) / (v * Sqr(T))
	// d2 = d1 - v * Sqr(T)
	double d1 = (log(F/data.K) + 0.5*v*v*data.T) / (v*sqrt(data.T));
	double d2 = d1 - v*sqrt(data.T);

	if (cp < 0)
	{ // Put
	
		// OutPerformance = Exp(-r * T) * (X * CND(-d2) - F * CND(-d1))
		return exp(-data.r*data.T)*(data.K*N(-d2) - F*N(-d1));
	}
	else 
	{ // Call
		
		//OutPerformance = Exp(-r * T) * (F * CND(d1) - X * CND(d2))
		return exp(-data.r*data.T)*((F*N(d1) - data.K*N(d2)));
	}
}

double ProductOption(const TwoFactorBasketData& data, double S1, double S2, int cp)
{
	
	/* F = S1 * S2 * Exp((b1 + b2 + rho * v1 * v2) * T)
    v = Sqr(v1 ^ 2 + v2 ^ 2 + 2 * rho * v1 * v2)
    d1 = (Log(F / X) + T * v ^ 2 / 2) / (v * Sqr(T))
    d2 = d1 - v * Sqr(T)

    If CallPutFlag = "c" Then
        ProductOption = Exp(-r * T) * (F * CND(d1) - X * CND(d2))
    Else
        ProductOption = Exp(-r * T) * (X * CND(-d2) - F * CND(-d1))
    End If*/

	// Haug
	double F = (S1*S2)*exp((data.b1 + data.b2 + data.rho*data.s1*data.s2)*data.T);

	double v = sqrt(data.s1*data.s1 + data.s2*data.s2 + 2.0*data.rho*data.s1*data.s2); 
	
	double d1 = (log(F/data.K) + 0.5*data.T*v*v)/ (v*sqrt(data.T));
	double d2 = d1 - v*sqrt(data.T);
	
	if (cp < 0)
	{ // Put

		return exp(-data.r*data.T)*(data.K*N(-d2) - F*N(-d1));
	}
	else 
	{ // Call

		return exp(-data.r*data.T)*(F*N(d1) - data.K*N(d2));
	}


	// Zhang
	/*double F = (S1*S2)*exp((data.r - data.q1 - data.q2 + data.rho*data.s1*data.s2)*data.T);

	double v = sqrt(data.s1*data.s1 + data.s2*data.s2 + 2.0*data.rho*data.s1*data.s2); 
	
	double d2 = (log(S1*S2/data.K) + (data.b1 + data.b2 - 0.5*(data.s1*data.s1 + data.s2*data.s2))*data.T)/ v*sqrt(data.T);
	double d1 = d2 + v*sqrt(data.T);
	//cout << d2 << ", " << d1 << ", " << v << ", " << data.T; int l; cin >> l;
	if (cp < 0)
	{ // Put

		return -F*N(-d1) + data.K*exp(-data.r*data.T)*N(-d2);
	}
	else 
	{ // Call

		return F*N(d1) - data.K*exp(-data.r*data.T)*N(d2);
	}*/

}

double SpreadOption(const TwoFactorBasketData& data, double S1, double S2, int cp)
{ // max(S1 - S2 - X, 0) or max(X - S1 - S2, 0)
	

	// Haug
	double F = data.Q2*S2*exp((data.b2 - data.r)*data.T) / (data.Q2*S2*exp((data.b2 - data.r)*data.T) + data.K*exp(-data.r*data.T));
	double S = data.Q1*S1*exp((data.b1 - data.r)*data.T) / (data.Q2*S2*exp((data.b2 - data.r)*data.T) + data.K*exp(-data.r*data.T));
	double sig = sqrt(data.s1*data.s1 + data.s2*data.s2*F*F - 2.0*data.rho*data.s1*data.s2*F);

	double d1 = (log(S) + 0.5*sig*sig*data.T) / (sig*sqrt(data.T));
	double d2 = d1 - sig*sqrt(data.T);

	if (cp < 0)
	{ // Put
		
		return (data.Q2*S2*exp((data.b2-data.r)*data.T) + data.K*exp(-data.r*data.T)) * (N(-d2) - S*N(-d1));
	}
	else 
	{ // Call
		
		return (data.Q2*S2*exp((data.b2-data.r)*data.T) + data.K*exp(-data.r*data.T)) * (S*N(d1) - N(d2));
	}
}


#endif
/*

int main()
{

	TwoFactorBasketData data;
	data.s1 = 0.2; data.s2 = 0.2; data.rho = 0.7; 
	data.b1 = 0.04; data.b2 = 0.06; // b = r - q for dividend paying stock
	data.T = 0.5;
	data.r = 0.1;
	data.Q1 = 1; data.Q2 = 1;

	data.s1 = 0.18; data.s2 = 0.12; data.rho = 0.8; 
	data.b1 = 0.02; data.b2 = 0.04;
	data.T = 0.5;
	data.r = 0.1;
	data.Q1 = 1; data.Q2 = 1;

	data.print();

	double S1 = 22.0; double S2 = 20.0;
	cout << "Exchange option price: " << Margrabe(data, S1, S2) << endl;

	S1 = 130.0; S2 = 100.0;
	data.s1 = 0.3; data.s2 = 0.4; data.rho = 0.0; 
	data.b1 = 0.05; data.b2 = 0.03; // b = r - q for dividend paying stock
	data.T = 0.5;
	data.r = 0.07; data.K = 0.5;
	cout << "Outperformance option put price: " << OutperformanceOption(data, S1, S2, -1) << endl;
	cout << "Outperformance option call price: " << OutperformanceOption(data, S1, S2, +1) << endl;

	return 0;
}
*/