// ExcelDriver.cpp
//
// Excel driver class.
//
// Datasim Education BV (c) 2003
// 
// Modification dates:
// 23 July 2003 - MB - Started.
// 29 July 2003 - MB - Added support for PropertySets.
//					 - Writing chart data in vertical direction.
//					 - Added Instance() method which returns an ExcelDriver singleton.
// 2005-10-16 DD Now to be changed for Intro C++ book
// 2005-12-1 DD hpp and cpp versions
// 2005-12-9 DD creatVector here
// 2005-12-10 DD MAJOR BUG FIXED, thanks to Martin Nisbit. Here is a 
// description of problem: 
// 2005-12-16 DD No exception at this level (AddMatrix)
// 2009-1-12 DD print matrix with sets as indices
// 2012-1-31 DD uBLAS
// 
/*
	Getting the Excel Driver to work is not quite trivial, and is covered on its own page. If you compile it successfully under Office XP or Office 2003, but get a "type mismatch" runtime error instead of results when trying to run it, you need to use this version instead. As Martin Nisbit (who provided that version) discovered, one has to replace all occurrences of "something.Value" by "Something.Value2" for the driver to work under these versions of Office, so for example instead of 
	item->Value = label.c_str();
	you have to use 
	item->Value2 = label.c_str();
	Also, the Excel Driver, even when it runs successfully, tends to crash upon exit - happened to me both when using Office 2000 and Office 2003, and Martin reports the same behavior under Office XP. When I will figure out what causes it, I'll write it here - or if you know, please write me!

*/
//
// (C) Datasim Education BV 2003 - 2006
//

#ifndef ExcelDriver_CPP
#define ExcelDriver_CPP

#include "ExcelDriver.hpp"
#include "StringConversions.hpp"
#include "DatasimException.hpp"
// Returns row from matrix as vector.
u::vector <double> createVector(const u::matrix<double>& mat, long row)
{

		u::vector<double> result (mat.size2());
		for (std::size_t i = 0; i < mat.size2(); i++)
		{
			result[i] = mat(row, i);
		}

		return result;
}

// Writes label and vector to cells in horizontal direction.
void ExcelDriver::ToSheetHorizontal( Excel::_WorksheetPtr sheet, 
					long sheetRow, long sheetColumn,
					const std::string& label, const u::vector<double> & values )
{
		// Get cells.
		Excel::RangePtr pRange=sheet->Cells;
	
		// First cell contains the label.
		Excel::RangePtr item = pRange->Item[sheetRow][sheetColumn];
	//	item->Value = label.c_str();
		item->Value2 = label.c_str();
	
		sheetColumn++;

		// Next cells contain the values.
		for(std::size_t i=0; i < values.size(); i++)
		{
			Excel::RangePtr item = pRange->Item[sheetRow][sheetColumn];
//			item->Value = values[i];			
			item->Value2 = values[i];			
	
			sheetColumn++;
		}
}

// Writes label and vector to cells in vertical direction.
void ExcelDriver::ToSheetVertical( Excel::_WorksheetPtr sheet,
					long sheetRow, long sheetColumn,
					const std::string& label, const u::vector<double> & values )
{
		// Get cells.
		Excel::RangePtr pRange=sheet->Cells;
	
		// First cell contains the label.
		Excel::RangePtr item = pRange->Item[sheetRow][sheetColumn];
		// item->Value = label.c_str();
		item->Value2 = label.c_str();
	
		sheetRow++;

		// Next cells contain the values.
		for(std::size_t i=0; i < values.size(); i++)
		{
			Excel::RangePtr item = pRange->Item[sheetRow][sheetColumn];
			//item->Value = values[i];			
			item->Value2 = values[i];
			sheetRow++;
		}
}

// Throw COM error as string.
void ExcelDriver::ThrowAsString( _com_error & error )
{
		bstr_t description = error.Description();
		if( !description )
		{
			description = error.ErrorMessage();
		}

		throw string(description);
}

// Constructor. Starts Excel in invisible mode.
ExcelDriver::ExcelDriver()	: curDataColumn(1)
{
		try
		{
			// Initialize COM Runtime Libraries.
			CoInitialize(NULL);

			// Start excel application.
			xl.CreateInstance(L"Excel.Application");
			xl->Workbooks->Add((long) Excel::xlWorksheet);
          //  xl->Workbooks->OpenFile("D:\Documents and Settings\dduffy\My Documents\Testje.xls");

			// Rename "Sheet1" to "Chart Data".
			Excel::_WorkbookPtr pWorkbook = xl->ActiveWorkbook;
			Excel::_WorksheetPtr pSheet = pWorkbook->Worksheets->GetItem("Sheet1");
			pSheet->Name = "Chart Data";
		}
		catch( _com_error & error )
		{
			ThrowAsString(error);
		}
}

// Destructor.
ExcelDriver::~ExcelDriver()
{
		// Undo initialization of COM Runtime Libraries.
		CoUninitialize();
	}

// Access to single, shared instance of ExcelDriver (singleton).
ExcelDriver& ExcelDriver::Instance()
{
		static ExcelDriver singleton;
		return singleton;
}


// Add Matrix to the spreadsheet with row and column labels.
void ExcelDriver::AddMatrix(const std::string& name, 
		const u::matrix<double> & matrix, 
		const list<std::string> & rowLabels, const list<std::string> & columnLabels)
	{
		
		// Add sheet.
		Excel::_WorkbookPtr pWorkbook = xl->ActiveWorkbook;
		Excel::_WorksheetPtr pSheet = pWorkbook->Worksheets->Add();
		pSheet->Name = name.c_str();

		// Current indeces in spreadsheet.
		long sheetRow = 1;
		long sheetColumn = 1;
				
		// Add column labels starting at column 2.
		sheetColumn = 2;
		Excel::RangePtr pRange=pSheet->Cells;
		list<string>::const_iterator columnLabelIt;
		for(columnLabelIt = columnLabels.begin(); columnLabelIt != columnLabels.end(); 
																		columnLabelIt++)
		{
			Excel::RangePtr item = pRange->Item[sheetRow][sheetColumn];
		//	item->Value = columnLabelIt->c_str();
			item->Value2 = columnLabelIt->c_str();
		
			sheetColumn++;
		}

		// Add row labels + values.
		sheetColumn = 1;
		sheetRow = 2;
		list<string>::const_iterator rowLabelIt = rowLabels.begin();
		for(std::size_t r = 0; r < matrix.size1(); r++ )
		{
			u::vector<double>	row = createVector(matrix, r);
			ToSheetHorizontal(pSheet, sheetRow, sheetColumn, *rowLabelIt, row);
			rowLabelIt++;
			sheetRow++;
		}
}


// Make Excel window visible.
void ExcelDriver::MakeVisible( bool b)
{
		// Make excel visible.
		xl->Visible = b ? VARIANT_TRUE : VARIANT_FALSE;
}

// For debugging, for example
void ExcelDriver::printStringInExcel(const string& s, long rowNumber, long colNumber,
					const string& sheetName	)
{ // Mainly for debugging

		// Add sheet.
		Excel::_WorkbookPtr pWorkbook = xl->ActiveWorkbook;
		Excel::_WorksheetPtr pSheet = pWorkbook->Worksheets->Add();
		pSheet->Name = sheetName.c_str();

		// Add properties to cells.
		Excel::RangePtr pRange = pSheet->Cells;
		pRange->Item[rowNumber][colNumber] = s.c_str();
}

void ExcelDriver::printStringInExcel(const list<string>& s, long startRow, long colNumber,
					const string& sheetName	)
{ // Mainly for debugging

		// Add sheet.
		Excel::_WorkbookPtr pWorkbook = xl->ActiveWorkbook;
		Excel::_WorksheetPtr pSheet = pWorkbook->Worksheets->Add();
//		pSheet->Name = sheetName.c_str();

		Excel::RangePtr pRange = pSheet->Cells;
	
		list<string>::const_iterator it;

		for (it = s.begin(); it != s.end(); it++)
		{
				
			pRange->Item[startRow][colNumber] = (*it).c_str();
			startRow++;
		}
}

// Print a two-dimensional array (typically, one time level)
void printMatrixInExcel(const u::matrix<double>& myMatrix,
				 const u::vector<double>& xarr, const u::vector<double>& yarr, 
				 const std::string& SheetName, bool show)
{

	// PREC: Type II, III
	if (xarr.size() != myMatrix.size1() || yarr.size() != myMatrix.size2())
	{
		//string tmp = "Vectors X, Y, matrix M: " + getString(xarr.size()) + comma + getString(yarr.size())
		//	+ bracketL + getString(matrix.Rows()) + comma + getString(matrix.Columns()) + bracketR;

		string tmp("KK");
		throw DatasimException(string("Size mismatch of vector(s) and matrix"), string("PrintMatrixInExcel"),
			tmp);
	}

	ExcelDriver& excel = ExcelDriver::Instance();
	excel.MakeVisible(show);
	// Print a matrix showing the x and y positions

	// We need to convert the double arrays to lists of strings
	cout << "Starting Excel";

	list<std::string> rowlabels = convertToString(xarr);
	list<std::string> columnlabels = convertToString(yarr);

	cout << "Staring Excel";
	
	excel.AddMatrix(SheetName, myMatrix, rowlabels, columnlabels);

}

#endif
