// TestTwoFactorPde.cpp
//
// Test programs for 2-factor BS using ADE. 
// 
// We use C++ 11 lambda functions to define payoffs; later we can bind a boost
// function to member function of a payoff class.
//
// (C) Datasim Education BV 2010-2012
//

#include "GeneralInclude.hpp"
#include "TwoFactorADE.hpp"
#include "ExactFormulae.hpp"


int main()
{
	// Choose the particular form of the PDE
	using namespace TwoFactorBlackScholesUnitTransformedPde;

	// Parameters of the BS PDE (2-asset)
	double S1; double S2; // Underlying values where is calculated

	// Haug 2007 (V = 0.8908)
/*	r = 0.07;
	s1 = 0.3; s2 = 0.4;
	q1 = 0.02; q2 = 0.04;
	T = 0.5; K = 0.5;
	rho = 0;
	//a1 = a2 = 1.0;
	S1 = 122; S2 = 120;	// Values where price is calculated
	*/


	// Haug page 206, exchange option
	r = 0.1;
	s1 = 0.2; s2 = 0.2;
	q1 = 0.06; q2 = 0.04;
	T = 1.0;
	rho = -0.5;
	S1 = 122; S2 = 220;	// Values where price is calculated
	double Q1 = 1.0; double Q2 = 1.0;
	double K = 0;// redundant
	payoff = [&] (double S1, double S2) {return max(Q1*S1 - Q2*S2, 0.0);};	
		
		
	// Haug page 214 spread option
/*	r = 0.1;
	s1 = 0.2; s2 = 0.2;
	q1 = r; q2 = r;
	T = 0.5;
	rho = -0.5;
	S1 = 122; S2 = 120;	// Values where price is calculated
	double K = 3.0;
	int cp = 1;
	if (cp > 0)
		payoff = [&] (double S1, double S2) {return max(S1 - S2 - K, 0.0);};		
	else
		payoff = [&] (double S1, double S2) {return max(K - S1 + S2, 0.0);};	
	*/
	// Haug page 203, relative outperformance option
/*	r = 0.07;
	s1 = 0.3; s2 = 0.4;
	q1 = 0.02; q2 = 0.04;
	T = 0.5;
	rho = 0.9;
	S1 = 130; S2 = 100;	// Values where price is calculated
	double K = 0.5;
	int cp = -1;
	if (cp > 0)
		payoff = [&] (double S1, double S2) -> double { if (S2 != 0.0) return max(S1/S2 - K, 0.0); else return 0.0;}; // call
	else
		payoff = [&] (double S1, double S2) -> double { if (S2 != 0.0) return max(K - S1/S2, 0.0); else return 0.0;}; // put
	*/
	// Haug page 205; product option
/*	r = 0.07;
	s1 = 0.2; s2 = 0.3;
	q1 = r - 0.02; q2 = r - 0.05;
	T = 0.5;
	rho = -0.9;
	S1 = 100; S2 = 105;	// Values where price is calculated
	double K = 15000.0;
	int cp = 1;
	if (cp > 0)
		payoff = [&] (double S1, double S2) -> double { return max(S1*S2 - K, 0.0);}; // call
	else
		payoff = [&] (double S1, double S2) -> double { return max(K - S1*S2, 0.0); }; // put
	*/	
		
	// Zhang, product option
/*	r = 0.05;
	s1 = 0.1; s2 = 0.15;
	q1 = 0.0; q2 = 0.0;

	T = 0.25;
	rho = 0.999;
	S1 = 4; S2 = 15;	// Values where price is calculated
	double K = 60.0;
	int cp = -1;
	if (cp > 0)
		payoff = [&] (double S1, double S2) -> double { return max(S1*S2 - K, 0.0);}; // call
	else
		payoff = [&] (double S1, double S2) -> double { return max(K - S1*S2, 0.0); }; // put
		*/
	// Domain transformation data; we use the transform z = x/(x + scale1),  w = y/(y + scale2). 
	// The the hotspot (S1, S2) in (x,y) space gets mapped to (1/2, 1/2) (usually) in (z,w) space.
	xMax = 1.0;						// x far field	
	yMax = 1.0;						// y FF

	double xHotSpot = 0.5; double yHotSpot = 0.5;
	scale1 = S1*(1.0 - xHotSpot)/xHotSpot; scale2 = S2*(1.0 - yHotSpot)/yHotSpot; 
		
	// Create pde and domain (factory object); it depends on the interface functions in the current
	// namespace (later do OO or other approach. We should make IBVP := PDE + PDEDomain.

	// N.B. The namespace contains the specific functionality for the PDE as well as factory methods.
	TwoFactorPde<double> pde;
	createTwoFactorPde(pde);
	TwoFactorPdeDomain<double> pdeDomain; createTwoFactorPdeDomain(pdeDomain);
	cout << "Now creating solver\n";

	// FD scheme
	long NXX = 100; long NYY = 100; long NTT =	100;
	cout << "Give NX (e.g. 200): "; cin >> NXX; 
	cout << "Give NY (e.g. 200): "; cin >> NYY; 
	cout << "Give NT (e.g. 200): "; cin >> NTT;
	
	
	TwoFactorADESolver solver(pde, pdeDomain, NXX, NYY, NTT);

	cout << "Now starting computation\n";

	// Run the state machine
	solver.result();
	u::matrix<double> m(solver.MatNew);
	
	cout << "Now starting Excel\n";
	bool show = true;

	try 
	{
	//	printMatrixInExcel(m, solver.xmesh, solver.ymesh, string("TWOFAC"), true);
	}
	catch(DatasimException& e)
	{
		e.print();

		ExcelDriver& excel = ExcelDriver::Instance();
		excel.MakeVisible(true);	
 
		excel.printStringInExcel(e.Message(), 1, 1, string("Err"));

		std::list<std::string> dump;
		dump.push_back(e.MessageDump()[0]);
		dump.push_back(e.MessageDump()[1]);
		dump.push_back(e.MessageDump()[2]);

		excel.printStringInExcel(dump, 1, 1, string("Err"));
	}
	
	// Exact solution; outperformance
	u::matrix<double> matExact(solver.xmesh.size(), solver.ymesh.size());

	TwoFactorBasketData data;
	data.s1 = s1; data.s2 = s2; data.rho = rho; 
	data.b1 = r - q1; data.b2 = r - q2; // b = r - q for dividend paying stock
	data.q1 = r - data.b1; data.q2 = r - data.b2; 
	data.T = T; data.r = r; 
	data.Q1 = 1; data.Q2 = 1;
	data.K = K;
	data.print();
	
	double SA, SB;
	for (std::size_t row = 0; row < matExact.size1(); ++row)
	{
		for (std::size_t column = 0; column < matExact.size2(); ++column)
		{
			SA = scale1*solver.xmesh[row]/(1.0 - solver.xmesh[row]);
			SB = scale2*solver.ymesh[column]/(1.0 - solver.ymesh[column]);
	
			//matExact(row, column) = OutperformanceOption(data, SA, SB, cp);
			//matExact(row, column) = ProductOption(data, SA, SB, cp);
			matExact(row, column) = Margrabe(data, SA, SB);
			//matExact(row, column) = SpreadOption(data, SA, SB, cp);
		}
	}

	
	// Postprocessing: choose the index ranges in which to analyse the error.


	// 1. Bounds for matrix ranges using STL
	double S1Max = 1.5*S1; double S2Max = 1.5*S2;
	double x1Max = S1Max/(S1Max + scale1); double y1Max = S2Max/(S2Max + scale2); 

	u::vector<double>::iterator pos1 = find_if(solver.xmesh.begin(), solver.xmesh.end(), [&] (double d) { return d >= x1Max; });
	u::vector<double>::iterator pos2 = find_if(solver.ymesh.begin(), solver.ymesh.end(), [&] (double d) { return d >= y1Max; });

	long iMax = distance(solver.xmesh.begin(), pos1) + 1;
	long jMax = distance(solver.ymesh.begin(), pos2) + 1;

	// 2. Create view of matrix using Boost
	u::matrix<double> matError = m - matExact;
	u::matrix_range<u::matrix<double> > mr1(matError, u::range(16, iMax), u::range(16,jMax));
	//u::matrix_range<u::matrix<double> > mr1(matError, u::range(4, matError.size1()), u::range(4,matError.size2()));

//printMatrixInExcel(matExact, solver.xmesh, solver.ymesh, string("Exact"), true);
//	printMatrixInExcel(matError, solver.xmesh, solver.ymesh, string("Error"), true);
	//printMatrixInExcel(mr1, solver.xmesh, solver.ymesh, string("Error2"));
	std::cout << "Max error, full matrix: " << u::norm_inf(mr1) << std::endl;
	
	// 3. Error at the hotspot (xHotSpot, yHotSpot)

	// Find index coordinates of hotspot
	u::vector<double>::iterator posA = find_if(solver.xmesh.begin(), solver.xmesh.end(),[&] (double d) { return d >= xHotSpot; } );
	u::vector<double>::iterator posB = find_if(solver.ymesh.begin(), solver.ymesh.end(),[&] (double d) { return d >= yHotSpot; } );

	long MaxA = distance(solver.xmesh.begin(), posA);
	long MaxB = distance(solver.ymesh.begin(), posB);

	cout << "\nApprox, Exact at hot spot: " << m(MaxA, MaxB) << ", " << matExact(MaxA, MaxB) << endl;
	// 4. Error in a box surrounding the hotspot

	u::vector<double>::iterator posABox = find_if(solver.xmesh.begin(), solver.xmesh.end(),[&] (double d) { return d >= xHotSpot; } );
	u::vector<double>::iterator posBBox = find_if(solver.ymesh.begin(), solver.ymesh.end(),[&] (double d) { return d >= yHotSpot; } );

	long MaxABox = distance(solver.xmesh.begin(), posABox);
	long MaxBBox = distance(solver.ymesh.begin(), posBBox);

	long width = 4;
	long off1 = width; long off2 = width; 
	long off3 = width; long off4 = width; 
	u::matrix_range<u::matrix<double> > mr2(matError, u::range(MaxABox - off1, MaxABox + off2), u::range(MaxBBox - off3, MaxBBox + off4));

	std::cout << "Max error, hotspot area: " << u::norm_inf(mr2) << std::endl;

	data.print();
	cout << OutperformanceOption(data, 130, 100, 1) << endl;
	//cout << ProductOption(data, 100, 105, 1);

	//double OutperformanceOption(double S1, double S2, int cp, double b1, double b2, double v1, double v2, double X, double T, double rho, double r)

/*	cout << OutperformanceOption(130,100, 1, 0.05, 0.03,0.3,0.4,0.5,0.5,0,.07);
	cout << "Spread option: " << SpreadOption(data, 122,120,1) << endl;
	data.rho = -0.5;
	cout << "Spread option: " << SpreadOption(data, 122,120,1);
	*/
	// 5. Remove far field, V1
	u::matrix<double> matView(mr1.size1(), mr1.size2());
	
	double max = 0.0;
	long ipos; long jpos;

	for (std::size_t j = 0; j < matView.size2(); ++j) 
	{
		for (std::size_t i = 0; i < matView.size1(); ++i) 
		{
			//matView(i,j) = fabs(mr1(i,j));	// Absolute values of differences
			matView(i,j) = mr1(i,j);			// Values of differences
			if (matView(i,j) > max)
			{
				max = matView(i,j);
				ipos = i;
				jpos = j;
			}
		}
	}
	
	cout << "\nmax " << max << ", at " << ipos << ", " << jpos << endl;

	u::vector<double> X(matView.size1()); u::vector<double> Y(matView.size2());
	for (std::size_t j = 0; j < X.size(); ++j)
	{
		X[j] = solver.xmesh[j];
	}
	for (std::size_t j = 0; j < Y.size(); ++j)
	{
		Y[j] = solver.ymesh[j];
	}

	printMatrixInExcel(matView, X, Y, string("Error"), show);

	return 0;
}