// TwoFactorADE.cpp
//
// Two-dimensional ADE solver. In this case we use one-sided approximations for the 
// convection and mixed derivative terms.
//
// 2010-7-8 DD code changed based on Heat 3d pde
// 2011-7-31 DD new update n -> n+1
// 2012-1-31 DD clean code
//
// (C) Datasim Education BV 2009-2012
//
#include "GeneralInclude.hpp"

#ifndef TwoFactorADESolver_hpp
#define TwoFactorADESolver_hpp

#include "TwoFactorPde.hpp"
#include "TwoFactorPdeDomain.hpp"
#include "mesher.hpp"

#include <iostream>
using namespace std;

//template <typename V>
class TwoFactorADESolver
{
private:
			TwoFactorPde<double> pde;				// The pde coefficients
			TwoFactorPdeDomain<double> pdeDomain;	// Domain, BC, IC

			// Data structures
			u::matrix<double> U;			// upper sweep, n+1
			u::matrix<double> UOld;		// upper sweep, n
			u::matrix<double> V;			// lower sweep, n+1
			u::matrix<double> VOld;		// lower sweep, n

public:
			u::matrix<double> MatNew;	// averaged solution, level n+1
			u::matrix<double> MatOld;	// averaged solution, level n
private:
			// Mesh-related data
			double hx, hy, delta_k, hx1, hy1, hx2, hy2;

			// Mesh-point values of coefficients
			double A,B,C,D,E,F,G,H;

			// Other variables
			double tprev, tnow, T;
			std::size_t NX, NY, NT;

public:
			u::vector<double> xmesh; 
			u::vector<double> ymesh; 
			u::vector<double> tmesh; 
private:
		
public:

		TwoFactorADESolver(TwoFactorPde<double>& source, TwoFactorPdeDomain<double>& domain,
					std::size_t NStepsX, std::size_t NStepsY, std::size_t NStepsT)
		{
			pde = source;
			pdeDomain = domain;

			NX = NStepsX;
			NY = NStepsY;
			NT = NStepsT;

			hx = pdeDomain.rx.spread() / double(NX);
			hy = pdeDomain.ry.spread() / double(NY);
			delta_k = pdeDomain.rt.spread() / double(NT);
	
			T = pdeDomain.rt.high();

			// Extra, handy variables
			hx1 = 1.0/hx;
			hx2 = 1.0/(hx*hx);
			hy1 = 1.0/hy;
			hy2 = 1.0/(hy*hy);

			// Create the mesh
			Mesher mesher(pdeDomain.rx,pdeDomain.ry);
			xmesh = mesher.xarr(NX); 
			ymesh = mesher.yarr(NY);

			// x == 1 correction
			xmesh[xmesh.size()-1] = 1.0 - 0.0001;
			ymesh[ymesh.size()-1] = 1.0 - 0.0001;

			tmesh = mesher.xarr(NT);
			tmesh[0] = pdeDomain.rt.low();
			tmesh[tmesh.size() - 1] = pdeDomain.rt.high();

			for (std::size_t n = 1; n < tmesh.size() - 1; ++n)
			{
				tmesh[n] = tmesh[n-1] + delta_k;
			}

		
			// Initialise U, UOld, V, VOld, MatNew data structures
			// NumericMatrix(I rows, I columns, I rowStart, I columnStart);		// Constructor with size & start index
			U = u::matrix<double>(xmesh.size(), ymesh.size());
			UOld = u::matrix<double>(xmesh.size(), ymesh.size());

			V = u::matrix<double>(xmesh.size(), ymesh.size());
			VOld = u::matrix<double>(xmesh.size(), ymesh.size());

			MatNew = u::matrix<double>(xmesh.size(), ymesh.size());
			MatOld = u::matrix<double>(xmesh.size(), ymesh.size());

			tprev = tnow = tmesh[0];
			initIC();

		}
			
		~TwoFactorADESolver()
		{
	
		}

		///////////////////////////////////////////////////////////////////
		void initIC()
		{ // Utility function to initialise the payoff function

		tprev = tnow = tmesh[0];

        // Initialise at the boundaries
		std::size_t lower = 0;
		std::size_t upper = ymesh.size() - 1;
		for (std::size_t i = 0; i < xmesh.size(); ++i)  
		{
			MatOld(i, lower) = pdeDomain.LowerBC(xmesh[i], tnow);
			MatOld(i, upper) = pdeDomain.UpperBC(xmesh[i], tnow);
		}

		// Left and Right BC
		std::size_t left = 0;
		std::size_t right = ymesh.size() - 1;
		for (std::size_t j = 0; j < ymesh.size(); ++j) 
		{
			MatOld(left, j) = pdeDomain.LeftBC(ymesh[j], tnow);
			MatOld(right, j) = pdeDomain.RightBC(ymesh[j], tnow);
		}
		
        // Now initialise values in interior of interval using
        // the initial function 'IC' from the PDE
      	for (std::size_t j = 0; j < ymesh.size(); ++j)
		{
			for (std::size_t i = 0; i < xmesh.size(); ++i) 
			{
				MatOld(i,j) = pdeDomain.IC(xmesh[i], ymesh[j]);
			}
		}

    }

		void calculateBC()
		{ // Calculate the discrete BC on the FOUR edges of the boundary

			// Lower and Upper BC
			std::size_t lower = 0;
			std::size_t upper = ymesh.size()-1;
			for (std::size_t i = 0; i < xmesh.size(); ++i)  
			{
				U(i, lower) = V(i, lower) = pdeDomain.LowerBC(xmesh[i], tnow);
				U(i, upper) = V(i, upper) = pdeDomain.UpperBC(xmesh[i], tnow);
			}

			// Left and Right BC
			std::size_t left = 0;
			std::size_t right = xmesh.size() -1;
			for (std::size_t j = 0; j < ymesh.size(); ++j) 
			{
				U(left, j) = V(left, j) = pdeDomain.LeftBC(ymesh[j], tnow);
				U(right, j) = V(right, j) = pdeDomain.RightBC(ymesh[j], tnow);
			}
		}

				
		void calculate()
		{ // Tells how to calculate sol. at n+1, Explicit ADE schemes	


				double mx, my;

				for (std::size_t j = 0; j < ymesh.size(); ++j) // Larkin
				{
					for (std::size_t i = 0; i < xmesh.size(); ++i) 
					{

							UOld(i,j) = MatOld(i,j);
							VOld(i,j) = MatOld(i,j);
					}
				}
				
				for (std::size_t j = 1; j < ymesh.size()-1; j++)
				{
					for (std::size_t i = 1; i < xmesh.size()-1; i++) 
					{

						// Create coefficients
						// hU_t = aU_xx + bU_yy + cU_xy + dU_x + eU_y + fU + G; // f = f(x,y,t)
						mx = xmesh[i];  my = ymesh[j];
						A = pde.a11(mx, my, tnow) * delta_k * hx2;
						B = pde.a22(mx, my, tnow) * delta_k * hy2;
						C = 0.25*pde.c(mx, my, tnow) * delta_k * hx1 * hy1;
						D = 0.5*pde.b1(mx, my, tnow) * delta_k * hx1;
						E = 0.5*pde.b2(mx, my, tnow) * delta_k * hy1;
						F = pde.d(mx, my, tnow) * delta_k;
						G = pde.F(mx, my, tnow) * delta_k;
				
						// Centred in convection, centred in mixed derivatives as well
						U(i, j) = (UOld(i,j)*(1.0-A-B) + UOld(i+1,j)*(A+D) + UOld(i,j+1)*(B+E) + U(i-1,j)*(A-D) + U(i,j-1)*(B-E) 
									+ C *(UOld(i+1,j+1) - UOld(i-1,j+1) - UOld(i+1,j-1) + UOld(i-1,j-1) ) // Janenko
									+ G) / (1.0+A+B-F);
					}
				}



				for (std::size_t j = ymesh.size()-2; j >= 1 ; j--)
				{
					for (std::size_t i = xmesh.size()-2; i >= 1; i--) 
					{
						// Create coefficients
						mx = xmesh[i];  my = ymesh[j];
						A = pde.a11(mx, my, tnow) * delta_k * hx2;
						B = pde.a22(mx, my, tnow) * delta_k * hy2;
						C = 0.25*pde.c(mx, my, tnow) * delta_k * hx1 * hy1; // V_xy
						D = 0.5*pde.b1(mx, my, tnow) * delta_k * hx1;
						E = 0.5*pde.b2(mx, my, tnow) * delta_k * hy1;
						F = pde.d(mx, my, tnow) * delta_k;
						G = pde.F(mx, my, tnow) * delta_k;
	
						// Centred in convection, centred in mixed derivatives as well
						V(i, j) = (VOld(i,j)*(1.0-A-B) + VOld(i-1,j)*(A-D) + VOld(i,j-1)*(B-E) + V(i+1,j)*(A+D) + V(i,j+1)*(B+E) 
									+ C *(VOld(i+1,j+1) - VOld(i-1,j+1) - VOld(i+1,j-1) + VOld(i-1,j-1))
									+ G) /(1.0+A+B-F);
					}
				}
			
		
			for (std::size_t j = 0; j < MatNew.size2() ; ++j)
			{
					for (std::size_t i = 0; i < MatNew.size1(); ++i) 
					{

							MatNew(i,j) = 0.5 * (U(i,j) + V(i,j));
							MatOld(i,j) = MatNew(i,j);
					}
			}

		
}


void result()
{ // The result of the calculation
	
	cout << "result";

	for (std::size_t n = 1; n < tmesh.size(); ++n)
	{
		if ((n/50)*50 == n)
		{
			cout << n << ", ";
		}

		tnow = tmesh[n];

		calculateBC(); // Calculate the BC at n+1
		calculate (); // Calculate the solution at n+1

		tprev = tnow;
	}

	// Finished at t = T; compare exact and discrete solutions

}

};
#endif