// TwoFactorBlackScholesUnitTransformedPde.hpp
//
// The 2-factor BS Pde to create the Pde and the Domain.
//
// 2011-8-1 DD Using IC as BC on boundaries.
// 2011-8-6 DD BS PDE transformed int [0,1]X[0,1]
// 2012-1-31 DD clean-up for qfcl
// 2012-2012 DD boost and lambda functions for customising Payoff functions
//
// (C) Datasim Education BC 2009-2012
//

#ifndef TwoFactorBlackScholesUnitTransformedPde_hpp
#define TwoFactorBlackScholesUnitTransformedPde_hpp

#include "TwoFactorPde.hpp"
#include "TwoFactorPdeDomain.hpp"
#include "GeneralInclude.hpp"
#include <cmath>
using namespace std;

namespace TwoFactorBlackScholesUnitTransformedPde
{

	// Scale factors (hotspots)
	double scale1;
	double scale2;
	
	// Parameters of PDE
	double r;
	double s1, s2, rho;
	double q1, q2;

	// Domain information
	double T;
	double xMax, yMax;

	
	double diffusionX(double x, double y, double t)
	{		
	
		return 0.5*s1*s1*x*x*(1.0-x)*(1.0-x);
	}

		
	double diffusionY(double x, double y, double t)
	{

		return 0.5*s2*s2*y*y*(1.0-y)*(1.0-y);
	}

	double convectionX(double x, double y, double t)
	{

		return (r - q1)*x*(1.0-x) - s1*s1*x*x*(1.0-x);
	}

	double convectionY(double x, double y, double t)
	{
		return (r - q2)*y*(1.0-y) - s2*s2*y*y*(1.0-y);
	}

	double termXY(double x, double y, double t)
	{
		return rho*s1*s2*x*y*(1.0-x)*(1.0-y);
	}

	double termFree(double x, double y, double t)
	{
		return -r;
	}

	double termInhomogeneous(double x, double y, double t)
	{
		return 0.0;
	}

	boost::function<double (double S1, double S2)> payoff;

	/*double payoff(double S1, double S2)
	{
		// Various 2 factor payoffs; hard-coded at the moment
		//return max(K - min(S1,S2), 0.0);
		//return max(K -S1 +S2, 0.0);		// put spread Haug 2007
		return max(S1 - S2 - K, 0.0);		// call spread Haug 2007
		//return max(S1+S2 - K, 0.0);
		//return max(0.0, S2-S1);
		//return max(max(-S1+110, -S2+100), 0.0); // Dual strike
		//return max(max(S1-110, -S2+100), 0.0); // Dual strike

		//return max(S1/S2 - K, 0.0);
	}
	*/

	double IC(double x, double y)
	{

		// Fudging at far field
		if (x == 1.0)
			x -= 0.01;

		if (y == 1.0)
			y -= 0.001;

		return payoff(scale1*x/(1.0-x), scale2*y/(1.0-y));
		
	}

	double BCLower(double x, double t)
	{
		return IC(x, 0.0)* exp(-r*t);
	}


	double BCUpper(double x, double t)
	{
		return IC(x, yMax);
	}

	double BCLeft(double y, double t)
	{ // y is S2
		
		return IC(0.0, y)* exp(-r*t);
	}
	
	double BCRight(double y, double t)
	{
		return IC(xMax, y);
	}

	// Factories
	void createTwoFactorPde(TwoFactorPde<double>& pde)
	{
		
		// Initialise all functions

		/*
		// U_t = a11U_Domain + a22U_yy + b1U_x + b2U_y + cU_xy + dU + F; // f = f(x,y,t)
		boost::function<T (T,T,T)> a11;
		boost::function<T (T,T,T)> a22;
		boost::function<T (T,T,T)> c;
		boost::function<T (T,T,T)> b1;
		boost::function<T (T,T,T)> b2;
		boost::function<T (T,T,T)> d;
		boost::function<T (T,T,T)> F;*/

		pde.a11 = &TwoFactorBlackScholesUnitTransformedPde::diffusionX;
		pde.a22 = &TwoFactorBlackScholesUnitTransformedPde::diffusionY;
		pde.c = &TwoFactorBlackScholesUnitTransformedPde::termXY;
		pde.b1 = &TwoFactorBlackScholesUnitTransformedPde::convectionX;
		pde.b2 = &TwoFactorBlackScholesUnitTransformedPde::convectionY;
		pde.d = &TwoFactorBlackScholesUnitTransformedPde::termFree;
		pde.F = &TwoFactorBlackScholesUnitTransformedPde::termInhomogeneous;

	}
	void createTwoFactorPdeDomain(TwoFactorPdeDomain<double>& pdeDomain)
	{
		
		pdeDomain.rx = Range<double>(0.0, xMax);
		pdeDomain.ry = Range<double>(0.0, yMax);
		pdeDomain.rt = Range<double>(0.0, T);

		pdeDomain.LeftBC = &TwoFactorBlackScholesUnitTransformedPde::BCLeft;
		pdeDomain.RightBC = &TwoFactorBlackScholesUnitTransformedPde::BCRight;
		pdeDomain.UpperBC = &TwoFactorBlackScholesUnitTransformedPde::BCUpper;
		pdeDomain.LowerBC = &TwoFactorBlackScholesUnitTransformedPde::BCLower;

		pdeDomain.IC = &IC;

	}
} // End of namespace

#endif