// TwoFactorExact.hpp
//
// Analytical solutions for various 2-factor option problems.
//
// (C) Datasim Education BV 2010

#ifndef TwoFactorExact_hpp
#define TwoFactorExact_hpp

#include <cmath>
#include "GeneralInclude.hpp"


double ExchangeCallOption(double S1, double S1, double t,					// hotspot
							double T, double q1, double q2,					// expiry and dividends
							double a1, double a2,							// max(a1*S1 - a2*S2, 0)
							double s1, double s2, double rho)				// vols and correlation
{

	double sNewSquared = s1*s1 - 2.0*rho*s1*s2 + s2*s2;
	double denom = sqrt(sNew)*sqrt(t);
	double tmp = log((a1*S1)/(as*S2));

	double d1 = (tmp + (q2 - q1 + 0.5*sNewSquared)*t);
	double d2 = (tmp + (q2 - q1 - 0.5*sNewSquared)*t);

	return a1*S1*exp(-q1*t)*N(d1) - a2*S2*exp(-q1*t)*N(d2);

}

double HullWhiteCallOption(double S, double t,					// hotspot
							double q, double a,					// from convection terms
							double s, double sr, double rho,	// volatility terms
							double P,							// price of zero-coupon bond
							double K)							// strike price
{

	double beta = sr*sr*(t + (2.0*exp(-a*t) - 0.5*exp(-2.0*a*t) - 1.5)/a) / a*a;
	double gamma = beta + s*s*t + 2.0*rho*sr*s*(t - (1.0 - exp(-a*t))/a) / a;
	gamma = sqrt(gamma);

	double tmp = log(S/(K*P));
	double d1 = (tmp - q*t + 0.5/(gamma*gamma)) / gamma;
	double d2 = (tmp - q*t - 0.5/(gamma*gamma)) / gamma;

	return S*exp(-q*t)*N(d1) - K*N(d2);
}

#endif		